#include <cstdio>
#include <fcntl.h>
#include <gtest/gtest.h>
#include <string>
#include <unistd.h>
#include <errors.h>
#include "disk_manager.h"

#define CREATE_FILE(d, path)						\
	DiskManager d;									\
	d.create_file(path);							\
	EXPECT_EQ(access(path.c_str(), F_OK), 0);

TEST(DiskManager, Default) {
	DiskManager disk_manager;
	EXPECT_EQ(disk_manager.MAX_FD, 8192);
}

TEST(DiskManager, CreateFile) {
	const std::string path = "test_create_file";
	CREATE_FILE(d, path);

	d.destroy_file(path);
}

TEST(DiskManager, IsFile) {
	const std::string path = "test_is_file", n_path = "test_not_file";
	CREATE_FILE(d, path);

	EXPECT_EQ(d.is_file(path), true);
	EXPECT_EQ(d.is_file(n_path), false);

	d.destroy_file(path);
}

TEST(DiskManager, DestroyFile) {
	const std::string path = "test_destroy_file";
	CREATE_FILE(d, path);

	EXPECT_EQ(access(path.c_str(), F_OK), 0);
	EXPECT_EQ(d.is_file(path), true);

	d.destroy_file(path);
	EXPECT_EQ(access(path.c_str(), F_OK), -1);
	EXPECT_EQ(d.is_file(path), false);
}

TEST(DiskManager, OpenFile) {
	DiskManager d;
	const std::string path = "test_open_file", n_path = "test_not_file";	

	// open before create
	EXPECT_THROW(d.open_file(path), FileNotFoundError);
	EXPECT_THROW(d.open_file(n_path), FileNotFoundError);

	d.create_file(path);

	// open first
	int fd = d.open_file(path);
	EXPECT_NE(fd, -1);
	d.close_file(fd);

	// open after close
	fd = d.open_file(path);
	EXPECT_NE(fd, -1);
	d.close_file(fd);

	d.destroy_file(path);

	// open after destroy
	EXPECT_THROW(d.open_file(path), FileNotFoundError);
	EXPECT_THROW(d.close_file(fd), FileNotOpenError);
}

TEST(DiskManager, SimpleTest) {
    srand((unsigned)time(nullptr));
	auto disk_manager = std::make_unique<DiskManager>();

	constexpr int MAX_FILES = 32;

    /** Test disk_manager */
    std::vector<std::string> filenames(MAX_FILES);  // MAX_FILES=32
    std::unordered_map<int, std::string> fd2name;
    for (size_t i = 0; i < filenames.size(); i++) {
        auto &filename = filenames[i];
        filename = std::to_string(i) + ".txt";
        if (disk_manager->is_file(filename)) {
            disk_manager->destroy_file(filename);
        }
        // open without create
        try {
            disk_manager->open_file(filename);
            EXPECT_TRUE(false);
        } catch (const FileNotFoundError &e) {
        }

        disk_manager->create_file(filename);
        EXPECT_TRUE(disk_manager->is_file(filename));
        try {
            disk_manager->create_file(filename);
            EXPECT_TRUE(false);
        } catch (const FileExistsError &e) {
        }

        // open file
        int fd = disk_manager->open_file(filename);

        fd2name[fd] = filename;

        disk_manager->set_fd2pageno(fd, 0);  // diskmanager在fd对应的文件中从0开始分配page_no
    }
}
